/* position.h
 */
#ifndef _POSITION_H
#define _POSITION_H

#include "osl/misc/loki.h"
#include "osl/offset.h"
#include <boost/utility/enable_if.hpp>

namespace osl 
{
  /**
   * 座標.
   *        盤面のインデックス
   * X, Yも1-9の範囲で表す 
   * Xは右から数える．Yは上から数える
   * なお駒台は0
   * <pre>
   * (A0)  ......................... (00)
   * (A1)  ......................... (01)
   * (A2) 92 82 72 62 52 42 32 22 12 (02)
   * (A3) 93 83 73 63 53 43 33 23 13 (03)
   * (A4) 94 84 74 64 54 44 34 24 14 (04)
   * (A5) 95 85 75 65 55 45 35 25 15 (05)
   * (A6) 96 86 76 66 56 46 36 26 16 (06)
   * (A7) 97 87 77 67 57 47 37 27 17 (07)
   * (A8) 98 88 78 68 58 48 38 28 18 (08)
   * (A9) 99 89 79 69 59 49 39 29 19 (09)
   * (AA) 9A 8A 7A 6A 5A 4A 3A 2A 1A (0A)
   * (AB) ...........................(0B)
   * (AC) ...........................(0C)
   * (AD) ...........................(0D)
   * (AE) ...........................(0E)
   * (AF) ...........................(0F) 
   * </pre>
   */
  class Position;
  bool operator==(Position l, Position r);

  class Position
  {
    unsigned int position;
    explicit Position(int p) : position(p)
    {
    }
  public:
    static const Position makeDirect(int value) { return Position(value); }
    unsigned int uintValue() const { return position; }
    enum {
      PIECE_STAND=0,
      MIN=0,
      SIZE=0x100
    };
    Position() : position(PIECE_STAND)
    {
    }
    static const Position STAND() { return Position(PIECE_STAND); }
    Position(int x, int y) : position((x*Offset::BOARD_HEIGHT)+y+1)
    {
      assert(position < SIZE);
    }
    /**
     * assertなしに作る
     */
    static const Position makeNoCheck(int x, int y) { 
      return Position((x*Offset::BOARD_HEIGHT)+y+1); 
    }
    static const Position nth(unsigned int i) { return Position(i+MIN); }
    /**
     * 将棋としてのX座標を返す. Positionの内部表現に依存しない．
     */
    int x() const { return position >> 4; }
    /**
     * 将棋としてのY座標を返す. Positionの内部表現に依存しない．
     */
    int y() const { return (position&0xf)-1; }
    /**
     * y+1を返す
     */
    int y1() const { return position&0xf; }
    unsigned int index() const { return position - MIN; }
    static unsigned int indexMax() { return SIZE - MIN; }
    int indexForOffset32() const { return position + (position&0xf0); }

    bool isPieceStand() const { return position == PIECE_STAND; }
    bool isOnBoardSlow() const;
    /**
     * 盤面上を表すかどうかの判定．
     * 1<=x() && x()<=9 && 1<=y() && y()<=9
     * Positionの内部表現に依存する．
     */
    bool isOnBoard() const { 
      return (0xffffff88&(position-0x12)&
	      ((unsigned int)((position&0x77)^0x12)+0xffffff77))==0;
    }
    /**
     * onBoardから8近傍のオフセットを足した点がedgeかどうかの判定
     * そこそこ速くなった．
     */
    bool isEdge() const { 
      assert(!isPieceStand() && 0<=x() && x()<=10 && 0<=y() && y()<=10);
      return (0x88&(position-0x12)&((position&0x11)+0xf7))!=0;
    }
    bool isValid() const;


    const Position positionForBlackSlow(Player player) const;

    const Position positionForBlack(Int2Type<BLACK>) const{
      assert(isOnBoard());
      Position ret=*this;
      return ret;
    }

    const Position positionForBlack(Int2Type<WHITE>) const{
      assert(isOnBoard());
      Position ret=makeDirect(Position(9,9).uintValue()+Position(1,1).uintValue()-uintValue());
      return ret;
    }
    /** 
     * 後手の場合は盤面を引っくり返す.
     * PIECE_STANDの場合は扱えない．
     */
    template<Player P>
    const Position positionForBlack() const{
      return positionForBlack(Int2Type<P>());
    }
    const Position positionForBlack(Player player) const{
      if(player==BLACK)
	return positionForBlack<BLACK>();
      else
	return positionForBlack<WHITE>();
    }

    const Position rotate180() const 
    {
      return positionForBlack<WHITE>();
    }
    const Position rotate180EdgeOK() const 
    {
      Position ret=makeDirect(Position(9,9).uintValue()+Position(1,1).uintValue()-uintValue());
      return ret;
    }
    const Position rotate180Safe() const 
    {
      if (isPieceStand())
	return *this;
      return positionForBlack<WHITE>();
    }
    const Position flipHorizontal() const
    {
      if (isPieceStand())
	return *this;
      return Position(10-x(), y());
    }

    static const Position onBoardMax(){ return Position(9,9); }
    static const Position onBoardMin(){ return Position(1,1); }
  
    /**
     * posがONBOARD_MINとONBOARD_MAXの間にある
     */
    bool isOnBoardRegion() const {
      return static_cast<unsigned int>(index()-onBoardMin().index()) 
	<= static_cast<unsigned int>(onBoardMax().index()-onBoardMin().index());
    }

    Position& operator++() {
      position += 1;
      return *this;
    }

    static int reverseX(int x) { return 10-x; }
    static int reverseY(int y) { return 10-y; }
  private:
    template<Player P>
    static bool canPromoteY(int y,Int2Type<P>);
    static bool canPromoteY(int y,Int2Type<BLACK>)
    {
      return y <= 3;
    }
    static bool canPromoteY(int y,Int2Type<WHITE>)
    {
      return y >= 7;
    }
    bool canPromote(Int2Type<BLACK>) const{
      return (uintValue()&0xf)<=4;
    }
    bool canPromote(Int2Type<WHITE>) const{
      return (uintValue()&0x8)!=0;
    }
  public:
    template <Player P>
    static bool canPromoteY(int y) { return canPromoteY(y,Int2Type<P>()); }
    template <Player P>
    bool canPromote() const{
      return canPromote(Int2Type<P>());
    }
    bool canPromote(Player player) const 
    {
      if (player==BLACK) 
	return canPromote<BLACK>();
      else 
	return canPromote<WHITE>();
    }
    /**
     * 2つのPosition(onBoardであることが前提)が，
     * xが等しいかyが等しい
     */
    bool isULRD(Position pos) const{
      assert(isOnBoard() && pos.isOnBoard());
      unsigned int v=uintValue() ^ pos.uintValue();
      return (((v+0xefull)^v)&0x110ull)!=0x110ull;
    }
    /**
     * 2つのPosition(onBoardであることが前提)のxが等しい
     */
    bool isUD(Position pos) const{
      assert(isOnBoard() && pos.isOnBoard());
      unsigned int v=uintValue() ^ pos.uintValue();
      return (v&0xf0)==0;
    }
    /**
     * posがPlayer Pにとって上
     */
    template<Player P>
    bool isU(Position pos) const{
      assert(isOnBoard() && pos.isOnBoard());
      unsigned int v=uintValue() ^ pos.uintValue();
      if(P==BLACK)
	return ((v|(uintValue()-pos.uintValue()))&0xf0)==0;
      else
	return ((v|(pos.uintValue()-uintValue()))&0xf0)==0;
    }
    /**
     * 2つのPosition(onBoardであることが前提)のyが等しい
     */
    bool isLR(Position pos) const{
      assert(isOnBoard() && pos.isOnBoard());
      unsigned int v=uintValue() ^ pos.uintValue();
      return (v&0xf)==0;
    }
    Position& operator+=(Offset offset) {
      position += offset.intValue();
      return *this;
    }
    Position& operator-=(Offset offset) {
      position -= offset.intValue();
      return *this;
    }
    const Position operator+(Offset offset) const {
      Position result(*this);
      return result+=offset;
    }
    const Position operator-(Offset offset) const {
      Position result(*this);
      return result-=offset;
    }
    const Offset operator-(Position other) const {
      return Offset::makeDirect(position - other.position);
    }
    template<int Y>
    bool yEq() {
      return (uintValue()&0xf)==(Y+1);
    }
    template<int Y>
    bool yLe(typename boost::enable_if_c<Y != 2>::type * =0) {
      return (uintValue()&0xf)<=(Y+1);
    }
    template<int Y>
    bool yLe(typename boost::enable_if_c<Y == 2>::type * =0) {
      return (uintValue()&0xc)==0;
    }
    template<int Y>
    bool yGe(typename boost::enable_if_c<Y!=7>::type * =0) {
      return (uintValue()&0xf)>=(Y+1);
    }
    template<int Y>
    bool yGe(typename boost::enable_if_c<Y==7>::type * =0) {
      return (uintValue()&0x8)!=0;
    }
  };

  inline bool operator==(Position l, Position r)
  {
    return l.uintValue() == r.uintValue();
  }
  inline bool operator!=(Position l, Position r)
  {
    return ! (l == r);
  }
  inline bool operator<(Position l, Position r)
  {
    return l.uintValue() < r.uintValue();
  }
  inline bool operator>(Position l, Position r)
  {
    return l.uintValue() > r.uintValue();
  }
  std::ostream& operator<<(std::ostream&, Position);

} // namespace osl

#endif /* _POSITION_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
